from django.contrib import admin

from .models import MessageAttachedUser


__all__ = 'MessageAttachedUserAdmin',


@admin.register(MessageAttachedUser)
class MessageAttachedUserAdmin(admin.ModelAdmin):
    list_display = 'id', 'message', 'user',
    list_select_related = 'message', 'user',
    autocomplete_fields = 'message', 'user',



import pytest

from zoneinfo import ZoneInfo
import datetime
from celery import Celery

from wcd_envoyer.contrib.scheduled.shortcuts import task, make_task, make_scheduled_task
from wcd_envoyer.models import ChannelConfig, Message
from wcd_envoyer.contrib.scheduled.models import (
    EventAvailability, ChannelAvailability, MessageSchedule
)

from .const import CONSOLE_CHANNEL
from .utils import make_template


EVENT_ONE = 'unexisting_event'


@pytest.fixture
@pytest.mark.django_db
def simple_console_config():
    config = ChannelConfig.objects.create(channel=CONSOLE_CHANNEL)
    return *make_template(CONSOLE_CHANNEL, EVENT_ONE), config


def make_send():
    app = Celery('app')
    app.conf.update(
        task_always_eager=True,
        broker_url='memory://',
        result_backend='cache+memory://',
    )
    runner_task = make_task(app)
    return runner_task.send, runner_task, app


@pytest.mark.django_db
def test_unscheduled_celery_send(django_assert_num_queries, simple_console_config, mocker):
    send, _, _ = make_send()

    with django_assert_num_queries(13):
        mocker.patch(
            'wcd_envoyer.channels.backends.console.ConsoleBackend.send',
            lambda self, config, messages, context: (messages, []),
        )
        send(
            EVENT_ONE,
            [{'user_id': 1}, {'user_id': 2}],
            {'username': 'Named', 'language': 'ja'},
        )

    assert Message.objects.all().count() == 2
    assert Message.objects.filter(status=Message.Status.SENT).count() == 2



@pytest.mark.django_db
def test_scheduled_celery_send(django_assert_num_queries, simple_console_config, mocker):
    send, _, _ = make_send()
    *_, config = simple_console_config
    UTC = ZoneInfo('UTC')
    now = datetime.datetime(2025, 1, 1, 1, 0, 0, 0, tzinfo=UTC)
    EventAvailability.objects.bulk_create([
        EventAvailability(
            event=EVENT_ONE,
            available_since=datetime.time(2, 0, 0),
            available_till=datetime.time(5, 0, 0),
        ),
    ])

    mocker.patch(
        'wcd_envoyer.channels.backends.console.ConsoleBackend.send',
        lambda self, config, messages, context: (messages, []),
    )
    send(
        EVENT_ONE,
        [{'user_id': 1}, {'user_id': 2}],
        {'username': 'Named', 'language': 'ja', 'now': now},
    )

    assert Message.objects.all().count() == 2
    assert Message.objects.filter(status=Message.Status.SENT).count() == 0
    assert Message.objects.filter(status=Message.Status.NEW).count() == 2
    assert MessageSchedule.objects.count() == 2
    assert MessageSchedule.objects.filter(send_at=now.replace(hour=2)).count() == 2


@pytest.mark.django_db
def test_intersecting_time_send(django_assert_num_queries, simple_console_config, mocker):
    send, _, _ = make_send()
    *_, config = simple_console_config
    UTC = ZoneInfo('UTC')
    now = datetime.datetime(2025, 1, 1, 1, 0, 0, 0, tzinfo=UTC)
    ChannelAvailability.objects.bulk_create([
        ChannelAvailability(
            config=config,
            available_since=datetime.time(1, 30, 0),
            available_till=datetime.time(2, 0, 0),
        ),
        ChannelAvailability(
            config=config,
            available_since=datetime.time(4, 30, 0),
            available_till=datetime.time(6, 0, 0),
        ),
    ])
    EventAvailability.objects.bulk_create([
        EventAvailability(
            event=EVENT_ONE,
            available_since=datetime.time(2, 0, 0),
            available_till=datetime.time(4, 0, 0),
        ),
        EventAvailability(
            event=EVENT_ONE,
            available_since=datetime.time(3, 0, 0),
            available_till=datetime.time(5, 0, 0),
        ),
    ])

    mocker.patch(
        'wcd_envoyer.channels.backends.console.ConsoleBackend.send',
        lambda self, config, messages, context: (messages, []),
    )
    send(
        EVENT_ONE,
        [{'user_id': 1}, {'user_id': 2}],
        {'username': 'Named', 'language': 'ja', 'now': now},
    )

    assert Message.objects.all().count() == 2
    assert Message.objects.filter(status=Message.Status.SENT).count() == 0
    assert Message.objects.filter(status=Message.Status.NEW).count() == 2
    assert MessageSchedule.objects.count() == 2
    assert MessageSchedule.objects.filter(send_at=now.replace(hour=2)).count() == 2


@pytest.mark.django_db
def test_send_tomorrow_because_late(django_assert_num_queries, simple_console_config, mocker):
    send, _, _ = make_send()
    *_, config = simple_console_config
    UTC = ZoneInfo('UTC')
    now = datetime.datetime(2025, 1, 1, 16, 0, 0, 0, tzinfo=UTC)
    ChannelAvailability.objects.bulk_create([
        ChannelAvailability(
            config=config,
            available_since=datetime.time(1, 30, 0),
            available_till=datetime.time(2, 30, 0),
        ),
    ])
    EventAvailability.objects.bulk_create([
        EventAvailability(
            event=EVENT_ONE,
            available_since=datetime.time(2, 0, 0),
            available_till=datetime.time(4, 0, 0),
        ),
    ])

    mocker.patch(
        'wcd_envoyer.channels.backends.console.ConsoleBackend.send',
        lambda self, config, messages, context: (messages, []),
    )
    send(
        EVENT_ONE,
        [{'user_id': 1}, {'user_id': 2}],
        {'username': 'Named', 'language': 'ja', 'now': now},
    )

    assert Message.objects.all().count() == 2
    assert Message.objects.filter(status=Message.Status.SENT).count() == 0
    assert Message.objects.filter(status=Message.Status.NEW).count() == 2
    assert MessageSchedule.objects.count() == 2
    assert MessageSchedule.objects.filter(send_at=now.replace(day=2, hour=2)).count() == 2


@pytest.mark.django_db
def test_send_other_timezone(django_assert_num_queries, simple_console_config, mocker):
    send, _, _ = make_send()
    *_, config = simple_console_config
    UTC = ZoneInfo('UTC')
    now = datetime.datetime(2025, 1, 1, 1, 0, 0, 0, tzinfo=UTC)
    ChannelAvailability.objects.bulk_create([
        ChannelAvailability(
            config=config,
            available_since=datetime.time(1, 30, 0),
            available_till=datetime.time(2, 30, 0),
        ),
    ])
    EventAvailability.objects.bulk_create([
        EventAvailability(
            event=EVENT_ONE,
            available_since=datetime.time(2, 0, 0),
            available_till=datetime.time(4, 0, 0),
        ),
    ])

    mocker.patch(
        'wcd_envoyer.channels.backends.console.ConsoleBackend.send',
        lambda self, config, messages, context: (messages, []),
    )
    send(
        EVENT_ONE,
        [
            {'user_id': 1, 'tz': 'Australia/West'},
            {'user_id': 2},
            {'user_id': 3, 'tz': 'Africa/Algiers'},
        ],
        {'username': 'Named', 'language': 'ja', 'now': now},
    )

    assert Message.objects.all().count() == 3
    assert Message.objects.filter(status=Message.Status.SENT).count() == 1
    assert Message.objects.filter(status=Message.Status.NEW).count() == 2
    assert MessageSchedule.objects.count() == 2
    assert MessageSchedule.objects.filter(send_at=now.replace(hour=2)).count() == 1
    assert MessageSchedule.objects.filter(send_at=now.replace(hour=18)).count() == 1


@pytest.mark.django_db
def test_scheduled_task(django_assert_num_queries, simple_console_config, mocker):
    send, t, app = make_send()
    scheduled_task = make_scheduled_task(app, sender=t.sender)

    *_, config = simple_console_config
    UTC = ZoneInfo('UTC')
    now = datetime.datetime(2025, 1, 1, 1, 0, 0, 0, tzinfo=UTC)
    ChannelAvailability.objects.bulk_create([
        ChannelAvailability(
            config=config,
            available_since=datetime.time(1, 30, 0),
            available_till=datetime.time(2, 30, 0),
        ),
    ])
    EventAvailability.objects.bulk_create([
        EventAvailability(
            event=EVENT_ONE,
            available_since=datetime.time(2, 0, 0),
            available_till=datetime.time(4, 0, 0),
        ),
    ])

    mocker.patch(
        'wcd_envoyer.channels.backends.console.ConsoleBackend.send',
        lambda self, config, messages, context: (messages, []),
    )
    send(
        EVENT_ONE,
        [
            {'user_id': 1, 'tz': 'Australia/West'},
        ],
        {'username': 'Named', 'language': 'ja', 'now': now},
    )

    assert Message.objects.all().count() == 1
    assert Message.objects.filter(status=Message.Status.NEW).count() == 1
    assert MessageSchedule.objects.count() == 1

    scheduled_task.apply_async(kwargs={
        'now': datetime.datetime(2025, 1, 1, 18, 0, 59, 0, tzinfo=UTC),
        'delta_seconds': datetime.timedelta(minutes=1).total_seconds(),
    })

    assert MessageSchedule.objects.count() == 0
    assert Message.objects.filter(status=Message.Status.NEW).count() == 0
    assert Message.objects.filter(status=Message.Status.SENT).count() == 1



from typing import *
from functools import partial, wraps
from wcd_envoyer.contrib.celery import shortcuts

from .services.sender import ScheduledCelerySender
from .tasks import actual_scheduled_sending_messages_task

if TYPE_CHECKING:
    from celery import Celery


__all__ = 'make_task', 'make_scheduled_task', 'task',


make_task = partial(shortcuts.make_task, sender_class=ScheduledCelerySender)
task = partial(shortcuts.task, sender_class=ScheduledCelerySender)


def make_scheduled_task(
    app: 'Celery',
    sender: ScheduledCelerySender,
    task_kwargs: dict = {},
    task_callback: Optional[Callable] = None,
    **kwargs,
):
    to_wrap = task_callback or actual_scheduled_sending_messages_task
    fn = wraps(to_wrap)(lambda *a, **kw: actual_scheduled_sending_messages_task(
        *a, **{**kwargs, **kw, 'sender': sender, 'callback': task_callback},
    ))
    return app.task(**task_kwargs)(fn)



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy


__all__ = 'AttachedUsersEnvoyerConfig',


class AttachedUsersEnvoyerConfig(AppConfig):
    name = 'wcd_envoyer.contrib.attached_users'
    label = 'wcd_envoyer_attached_users'
    verbose_name = pgettext_lazy('wcd_envoyer', 'Envoyer: Attached users')



from django.db import models
from django.utils.translation import pgettext_lazy

from wcd_envoyer import channels, events
from wcd_envoyer.models import Message, ChannelConfig
from wcd_envoyer.utils import DynamicChoicesField


__all__ = 'MessageSchedule', 'ChannelAvailability', 'EventAvailability',


class MessageSchedule(models.Model):
    class Meta:
        verbose_name = pgettext_lazy('wcd_envoyer', 'Message schedule')
        verbose_name_plural = pgettext_lazy('wcd_envoyer', 'Message schedules')

    message = models.OneToOneField(
        Message, on_delete=models.CASCADE,
        verbose_name=pgettext_lazy('wcd_envoyer', 'Message'),
        related_name='schedule',
    )
    send_at = models.DateTimeField(
        pgettext_lazy('wcd_envoyer', 'Send at'), null=False, blank=False,
    )

    def __str__(self) -> str:
        return pgettext_lazy('wcd_envoyer', '#{id}: {send_at}').format(
            id=self.message_id, send_at=self.send_at.isoformat(),
        )


class ChannelAvailability(models.Model):
    class Meta:
        verbose_name = pgettext_lazy('wcd_envoyer', 'Channel availability')
        verbose_name_plural = pgettext_lazy('wcd_envoyer', 'Channel availabilities')

    config = models.ForeignKey(
        ChannelConfig, on_delete=models.CASCADE,
        verbose_name=pgettext_lazy('wcd_envoyer', 'Channel config'),
        related_name='channel_availability',
    )
    available_since = models.TimeField(
        pgettext_lazy('wcd_envoyer', 'Available since'),
        null=False, blank=False,
    )
    available_till = models.TimeField(
        pgettext_lazy('wcd_envoyer', 'Available till'),
        null=False, blank=False,
    )

    def __str__(self):
        return f'#{self.id}'


class EventAvailability(models.Model):
    class Meta:
        verbose_name = pgettext_lazy('wcd_envoyer', 'Event availability')
        verbose_name_plural = pgettext_lazy('wcd_envoyer', 'Event availabilities')

    event = DynamicChoicesField(
        pgettext_lazy('wcd_envoyer', 'Event'),
        max_length=512, choices=events.registry.get_choices,
    )
    available_since = models.TimeField(
        pgettext_lazy('wcd_envoyer', 'Available since'),
        null=False, blank=False,
    )
    available_till = models.TimeField(
        pgettext_lazy('wcd_envoyer', 'Available till'),
        null=False, blank=False,
    )

    def __str__(self):
        return f'#{self.id}'



from typing import *
from datetime import datetime
from collections import defaultdict

from django.utils import timezone
from django.db.models import prefetch_related_objects

from wcd_envoyer.channels.backend import BaseMessagingBackend
from wcd_envoyer.services.sender import TemplatedMessagesGroup
from wcd_envoyer.models import ChannelConfig, Message
from wcd_envoyer.contrib.celery.services.sender import CelerySender
from wcd_envoyer.channels import MessageData

from . import scheduler
from ..models import MessageSchedule, EventAvailability


__all__ = 'ScheduledSenderMixin', 'ScheduledCelerySender',


MadeMessagesResult = List[Tuple[BaseMessagingBackend, ChannelConfig, List[Message]]]


class ScheduledSenderMixin:
    NOW_KEY: str = 'now'
    SCHEDULED_KEY: str = '_id_scheduled'
    IMMEDIATE_KEY: str = scheduler.IMMEDIATE_KEY
    TZ_KEY: str = scheduler.TZ_KEY

    def get_channel_configs(self, channels: Sequence[str]):
        configs = super().get_channel_configs(channels)
        prefetch_related_objects(list(configs.values()), 'channel_availability')

        return configs

    def get_event_availability_configs(self):
        availabilities = EventAvailability.objects.all()
        result = defaultdict(list)

        for availability in availabilities:
            result[availability.event].append(availability)

        return result

    def send_messages(self, messages: Iterable[Message], context: dict = {}):
        succeeded, failed = super().send_messages(messages, context=context)
        succeeded_pks = [y.pk for _, _, x in succeeded for y in x]

        if len(succeeded_pks) > 0:
            MessageSchedule.objects.filter(message_id__in=succeeded_pks).delete()

        return succeeded, failed

    def is_message_scheduled(self, message: Message) -> bool:
        return getattr(message, self.SCHEDULED_KEY, False)

    def split_scheduled_messages(
        self,
        templated_messages: Iterable[TemplatedMessagesGroup],
        context: dict = {},
    ) -> Dict[Optional[datetime], List[TemplatedMessagesGroup]]:
        if getattr(context, self.IMMEDIATE_KEY, False):
            return {None: list(templated_messages)}

        now = context.get(self.NOW_KEY) or timezone.now()
        tz_key = self.TZ_KEY
        splitted = defaultdict(list)
        availability_configs = self.get_event_availability_configs()

        for backend, config, items in templated_messages:
            splitted_items = scheduler.split_scheduled_messages(
                list(config.channel_availability.all()),
                items, now, tz_key=tz_key, immediate_key=self.IMMEDIATE_KEY,
                events_availability=availability_configs,
            )

            for schedule_at, messages in splitted_items.items():
                splitted[schedule_at].append((backend, config, messages))

        splitted[None] = splitted.pop(now, [])

        return splitted

    def make_messages(
        self,
        templated_messages: Iterable[TemplatedMessagesGroup],
        context: dict = {},
    ) -> MadeMessagesResult:
        splitted = self.split_scheduled_messages(
            templated_messages, context=context,
        )
        result = defaultdict(list)
        scheduled = []

        for send_at, data in splitted.items():
            for backend, config, items in data:
                messages = self.make_backend_messages(
                    backend, config, items, context=context,
                )

                if len(messages) > 0:
                    result[(backend, config)] += messages

                    if send_at is not None:
                        scheduled.append((send_at, messages))

        all_messages = Message.objects.bulk_create([
            item
            for items in result.values()
            for item in items
        ])

        if len(scheduled) > 0:
            schedules = MessageSchedule.objects.bulk_create([
                MessageSchedule(message_id=message.pk, send_at=send_at)
                for send_at, messages in scheduled
                for message in messages
            ])
            messages_map = {m.pk: m for m in all_messages}
            set_key = self.SCHEDULED_KEY

            for schedule in schedules:
                message = messages_map[schedule.message_id]
                setattr(message, set_key, True)
                message.schedule = schedule

        return [
            (backend, config, items)
            for (backend, config), items in result.items()
        ]


class ScheduledCelerySender(ScheduledSenderMixin, CelerySender):
    def send(
        self,
        messages_data: Sequence[MessageData],
        context: dict = {},
        only_channels: Optional[Sequence[str]] = None,
    ):
        messages_groups = self.prepare_messages(
            messages_data, context=context, only_channels=only_channels,
        )

        if len(messages_groups) != 0:
            self.send_messages_through_celery(
                [
                    y.pk
                    for _, _, x in messages_groups
                    for y in x
                    if not self.is_message_scheduled(y)
                ],
                context,
            )

        return [], []



from typing import *
from datetime import datetime, timedelta

from django.utils import timezone
from wcd_envoyer.models import Message

from .services.sender import ScheduledCelerySender
from .models import MessageSchedule


__all__ = 'actual_scheduled_sending_messages_task',


def actual_scheduled_sending_messages_task(
    sender: ScheduledCelerySender,
    now: Optional[datetime] = None,
    delta_seconds: float = timedelta(hours=1).total_seconds(),
    context: dict = {},
    only_new: bool = True,
    callback: Optional[Callable] = None,
):
    now = timezone.now() if now is None else now
    delta = timedelta(seconds=delta_seconds)
    since = now - delta
    messages = Message.objects.filter(pk__in=MessageSchedule.objects.filter(
        send_at__gte=since, send_at__lte=now,
    ).values('message_id'))

    if only_new:
        messages = messages.filter(status=Message.Status.NEW)

    result = sender.send_messages(messages, context=context)

    if callback is not None:
        callback(result, context)



from typing import *

from wcd_envoyer.contrib.celery.services.sender import CelerySender
from wcd_envoyer.models import Message


__all__ = 'actual_sending_messages_task',


def actual_sending_messages_task(
    sender: CelerySender,
    message_ids: List[int] = [],
    context: dict = {},
    only_pending: bool = False,
    callback: Optional[Callable] = None,
):
    messages = Message.objects.filter(pk__in=message_ids)

    if only_pending:
        messages = messages.filter(status=Message.Status.PENDING)

    result = sender.send_messages(messages, context=context)

    if callback is not None:
        callback(result, context)



from typing import *

from zoneinfo import ZoneInfo
from django.db import models
from django.core.serializers.json import DjangoJSONEncoder

from .functional import autoimport


__all__ = 'get_json_encoder', 'EnvoyerJSONEncoder',


def get_json_encoder():
    from ..conf import settings

    return autoimport(settings.JSON_ENCODER)


class EnvoyerJSONEncoder(DjangoJSONEncoder):
    def default(self, o):
        if isinstance(o, ZoneInfo):
            return o.key
        elif isinstance(o, models.Model):
            from django.contrib.contenttypes.models import ContentType

            return {
                'content_type': ContentType.objects.get_for_model(o.__class__),
                'pk': o.pk,
                'display': str(o),
            }
        elif isinstance(o, Exception):
            return str(o)
        else:
            return super().default(o)



from django.contrib import admin
from django.contrib.admin.filters import EmptyFieldListFilter
from django.contrib.admin.sites import site as admin_site
from django.utils.translation import pgettext_lazy

from wcd_envoyer.admin import ChannelConfigAdmin, MessageAdmin as MessageAdminInitial
from wcd_envoyer.models import Message

from .models import MessageSchedule, ChannelAvailability, EventAvailability


class ChannelAvailabilityInline(admin.StackedInline):
    model = ChannelAvailability
    extra = 0


ChannelConfigAdmin.inlines = list(ChannelConfigAdmin.inlines) + [
    ChannelAvailabilityInline,
]


@admin.register(EventAvailability)
class EventAvailabilityAdmin(admin.ModelAdmin):
    list_display = 'id', 'event', 'available_since', 'available_till',
    list_editable = 'event', 'available_since', 'available_till',
    list_filter = 'event',



MessageAdminBase = MessageAdminInitial

if admin_site.is_registered(Message):
    MessageAdminBase = admin_site._registry[Message].__class__
    admin_site.unregister(Message)


@admin.register(Message)
class MessageAdmin(MessageAdminBase):
    list_select_related = list(MessageAdminBase.list_select_related or []) + [
        'schedule',
    ]
    list_display = list(MessageAdminBase.list_display) + ['get_schedule_send_at']
    list_filter = list(MessageAdminBase.list_filter) + [
        ('schedule', EmptyFieldListFilter),
    ]

    def get_schedule_send_at(self, obj):
        return obj.schedule.send_at
    get_schedule_send_at.short_description = pgettext_lazy('wcd_envoyer', 'Send at')
    get_schedule_send_at.admin_order_field = 'schedule__send_at'



# Generated by Django 4.2.6 on 2024-01-29 09:12

from django.db import migrations, models
import django.db.models.deletion
import wcd_envoyer.utils.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wcd_envoyer', '0002_auto_20230505_0852'),
    ]

    operations = [
        migrations.CreateModel(
            name='EventAvailability',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('event', wcd_envoyer.utils.models.DynamicChoicesField(choices=[], max_length=512, verbose_name='Event')),
                ('available_since', models.TimeField(verbose_name='Available since')),
                ('available_till', models.TimeField(verbose_name='Available till')),
            ],
            options={
                'verbose_name': 'Event availability',
                'verbose_name_plural': 'Event availabilities',
            },
        ),
        migrations.CreateModel(
            name='MessageSchedule',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('send_at', models.DateTimeField(verbose_name='Send at')),
                ('message', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='schedule', to='wcd_envoyer.message', verbose_name='Message')),
            ],
            options={
                'verbose_name': 'Message schedule',
                'verbose_name_plural': 'Message schedules',
            },
        ),
        migrations.CreateModel(
            name='ChannelAvailability',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('available_since', models.TimeField(verbose_name='Available since')),
                ('available_till', models.TimeField(verbose_name='Available till')),
                ('config', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='channel_availability', to='wcd_envoyer.channelconfig', verbose_name='Channel config')),
            ],
            options={
                'verbose_name': 'Channel availability',
                'verbose_name_plural': 'Channel availabilities',
            },
        ),
    ]



from typing import *
from datetime import datetime
from collections import defaultdict
from django.contrib.auth import get_user_model

from django.utils import timezone
from django.db.models import prefetch_related_objects

from wcd_envoyer.channels.backend import BaseMessagingBackend
from wcd_envoyer.services.sender import TemplatedMessagesGroup, TemplatedMessages
from wcd_envoyer.models import ChannelConfig, Message
from wcd_envoyer.contrib.celery.services.sender import CelerySender
from wcd_envoyer.channels import MessageData

from ..models import MessageAttachedUser


__all__ = 'AttachedUserSenderMixin',


class AttachedUserSenderMixin:
    USER_ID_KEY: str = 'user_id'
    ATTACHED_USERS_KEY: str = 'attached_user_identifiers'

    def inject_user_ids_to_context(self, items: TemplatedMessages):
        user_id_key = self.USER_ID_KEY
        users_key = self.ATTACHED_USERS_KEY

        for message, _ in items:
            user_identifiers = {
                recipient.get(user_id_key)
                for recipient in message.recipients
            } - {None}

            if len(user_identifiers) == 0:
                continue

            message.context = {
                **message.context, users_key: list(user_identifiers),
            }

        return items

    def attach_users(self, messages: Iterable[Message]):
        attaches = []
        users_key = self.ATTACHED_USERS_KEY

        for message in messages:
            for user_id in message.data.get(users_key, []):
                attaches.append(MessageAttachedUser(
                    message_id=message.pk, user_id=user_id,
                ))

        existing_users = set(
            get_user_model().objects
            .filter(pk__in=[x.user_id for x in attaches])
            .values_list('pk', flat=True)
        )

        MessageAttachedUser.objects.bulk_create(
            [x for x in attaches if x.user_id in existing_users],
            ignore_conflicts=True,
        )

        return attaches

    def make_backend_messages(
        self,
        backend: BaseMessagingBackend,
        config: ChannelConfig,
        items: TemplatedMessages,
        context: dict = {},
    ) -> List[Message]:
        return backend.make_messages(
            config, self.inject_user_ids_to_context(items), context=context,
        )

    def make_messages(
        self,
        templated_messages: Iterable[TemplatedMessagesGroup],
        context: dict = {},
    ) -> List[Tuple[BaseMessagingBackend, ChannelConfig, List[Message]]]:
        result = super().make_messages(templated_messages, context=context)

        self.attach_users((y for _, _, x in result for y in x))

        return result



# Generated by Django 4.2.6 on 2024-01-30 07:03

from django.db import migrations, models
import wcd_envoyer.utils.types


class Migration(migrations.Migration):

    dependencies = [
        ('wcd_envoyer', '0002_auto_20230505_0852'),
    ]

    operations = [
        migrations.AlterField(
            model_name='channelconfig',
            name='data',
            field=models.JSONField(blank=True, default=dict, encoder=wcd_envoyer.utils.types.EnvoyerJSONEncoder, verbose_name='Data'),
        ),
        migrations.AlterField(
            model_name='channelconfig',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='message',
            name='data',
            field=models.JSONField(blank=True, default=dict, encoder=wcd_envoyer.utils.types.EnvoyerJSONEncoder, verbose_name='Data'),
        ),
        migrations.AlterField(
            model_name='message',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='message',
            name='recipients',
            field=models.JSONField(default=list, verbose_name='Recipients'),
        ),
        migrations.AlterField(
            model_name='template',
            name='data',
            field=models.JSONField(blank=True, default=dict, encoder=wcd_envoyer.utils.types.EnvoyerJSONEncoder, verbose_name='Data'),
        ),
        migrations.AlterField(
            model_name='template',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='templatetranslation',
            name='data',
            field=models.JSONField(blank=True, default=dict, encoder=wcd_envoyer.utils.types.EnvoyerJSONEncoder, verbose_name='Data'),
        ),
        migrations.AlterField(
            model_name='templatetranslation',
            name='language',
            field=models.CharField(choices=[('ru', 'Russian'), ('uk', 'Ukrainian'), ('pl', 'Polish'), ('cs', 'Czech'), ('ru-md', 'Russian-md'), ('ru-uz', 'Russian-uz'), ('cs-sk', 'Czech-sk'), ('ro', 'Romanian'), ('sk', 'Slovak'), ('en', 'English')], default='ru', max_length=32, verbose_name='Language'),
        ),
    ]



import pytest

from zoneinfo import ZoneInfo
import datetime
from celery import Celery

from django.contrib.auth import get_user_model
from wcd_envoyer.contrib.scheduled.shortcuts import make_task
from wcd_envoyer.contrib.scheduled.services.sender import ScheduledCelerySender
from wcd_envoyer.contrib.attached_users.services.sender import AttachedUserSenderMixin
from wcd_envoyer.models import ChannelConfig, Message
from wcd_envoyer.contrib.attached_users.models import MessageAttachedUser
from wcd_envoyer.contrib.scheduled.models import (
    EventAvailability, ChannelAvailability,
)

from .const import CONSOLE_CHANNEL
from .utils import make_template


EVENT_ONE = 'unexisting_event'


class Sender(AttachedUserSenderMixin, ScheduledCelerySender):
    pass


@pytest.fixture
@pytest.mark.django_db
def simple_console_config():
    config = ChannelConfig.objects.create(channel=CONSOLE_CHANNEL)
    return *make_template(CONSOLE_CHANNEL, EVENT_ONE), config


def make_send():
    app = Celery('app')
    app.conf.update(
        task_always_eager=True,
        broker_url='memory://',
        result_backend='cache+memory://',
    )
    runner_task = make_task(app, sender_class=Sender)
    return runner_task.send, runner_task, app


@pytest.mark.django_db
def test_existing_users_and_not(django_assert_num_queries, simple_console_config, mocker):
    send, _, _ = make_send()
    *_, config = simple_console_config
    UTC = ZoneInfo('UTC')
    now = datetime.datetime(2025, 1, 1, 1, 0, 0, 0, tzinfo=UTC)
    ChannelAvailability.objects.bulk_create([
        ChannelAvailability(
            config=config,
            available_since=datetime.time(1, 30, 0),
            available_till=datetime.time(2, 30, 0),
        ),
    ])
    EventAvailability.objects.bulk_create([
        EventAvailability(
            event=EVENT_ONE,
            available_since=datetime.time(2, 0, 0),
            available_till=datetime.time(4, 0, 0),
        ),
    ])
    User = get_user_model()
    user = User.objects.create_user('login', 'email@email.com', 'pass')

    mocker.patch(
        'wcd_envoyer.channels.backends.console.ConsoleBackend.send',
        lambda self, config, messages, context: (messages, []),
    )
    send(
        EVENT_ONE,
        [
            {'user_id': user.pk, 'tz': 'Australia/West'},
            {'user_id': 2},
            {'user_id': 3, 'tz': 'Africa/Algiers'},
        ],
        {'username': 'Named', 'language': 'ja', 'now': now},
    )

    assert Message.objects.all().count() == 3
    assert Message.objects.filter(status=Message.Status.SENT).count() == 1
    assert Message.objects.filter(status=Message.Status.NEW).count() == 2
    assert MessageAttachedUser.objects.count() == 1
    assert MessageAttachedUser.objects.filter(user_id=user.pk).count() == 1




# Generated by Django 4.2.6 on 2024-01-30 07:03

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wcd_envoyer', '0003_alter_channelconfig_data_alter_channelconfig_id_and_more'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='MessageAttachedUser',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('message', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attached_users', to='wcd_envoyer.message', verbose_name='Message')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='envoyer_messages', to=settings.AUTH_USER_MODEL, verbose_name='Send at')),
            ],
            options={
                'verbose_name': 'Message schedule',
                'verbose_name_plural': 'Message schedules',
            },
        ),
    ]



from typing import *
from dataclasses import dataclass, field

from .utils import EntryRegistry


__all__ = 'EventInfoType', 'EventDescriptor', 'EventRegistry',

ContextKey = Tuple[str, str, str]
ContextKeyInfo = Union[Tuple[str, str], ContextKey, str]


class EventInfoType(TypedDict):
    key: str
    verbose_name: Optional[str]
    context: Sequence[ContextKeyInfo]


def clear_context_key(key: ContextKeyInfo) -> ContextKey:
    if isinstance(key, str):
        return (key, key.title(), '')

    key, verbose_name, help_text, *_ = tuple(key) + ('', '')

    return (key, verbose_name if verbose_name else key.title(), help_text)


@dataclass
class EventDescriptor:
    key: str
    verbose_name: str
    context: List[ContextKey] = field(default_factory=list)

    @classmethod
    def from_info(cls, entry: EventInfoType):
        key = entry['key']
        verbose_name = entry.get('verbose_name', None) or key.title()

        return cls(
            key=key,
            verbose_name=verbose_name,
            context=[
                clear_context_key(key)
                for key in entry.get('context', [])
            ],
        )


class EventRegistry(EntryRegistry):
    def transform_input(self, entry: EventInfoType):
        return entry['key'], EventDescriptor.from_info(entry)

    def get_choices(self):
        return [(key, x.verbose_name) for key, x in self.items()]


registry = EventRegistry()
add = registry.add



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy


__all__ = 'ScheduledEnvoyerConfig',


class ScheduledEnvoyerConfig(AppConfig):
    name = 'wcd_envoyer.contrib.scheduled'
    label = 'wcd_envoyer_scheduled'
    verbose_name = pgettext_lazy('wcd_envoyer', 'Envoyer: Scheduled')



from typing import *
from collections import deque
from zoneinfo import ZoneInfo

from collections import defaultdict
from datetime import datetime, time, timezone, timedelta, tzinfo
from wcd_envoyer.channels import TemplatedMessages, MessageData

from ..models import ChannelAvailability, EventAvailability


__all__ = 'TZ_KEY', 'IMMEDIATE_KEY', 'split_scheduled_messages',


TimeRange = Tuple[time, time]
TimeRanges = List[TimeRange]
TZ_KEY = 'tz'
IMMEDIATE_KEY = 'send_immediate'


def resolve_tz(tz: Optional[Union[str, tzinfo]]) -> Optional[tzinfo]:
    if tz is None or isinstance(tz, tzinfo):
        return tz

    if isinstance(tz, str):
        return ZoneInfo(tz)


def normalize_ranges(ranges: Iterable[TimeRange]) -> TimeRanges:
    result = []
    q = deque(sorted(ranges))

    while q:
        f_since, f_till = q.popleft()

        while q:
            s_since, s_till = nxt = q.popleft()

            if not (f_till >= s_since and f_since <= s_till):
                q.appendleft(nxt)
                break

            f_since = min(f_since, s_since)
            f_till = max(f_till, s_till)

        result.append((f_since, f_till))

    return result


def intersect_timeranges(first: TimeRanges, second: TimeRanges) -> TimeRanges:
    if len(first) == 0:
        return second

    if len(second) == 0:
        return first

    intersects = []

    for f_since, f_till in first:
        for s_since, s_till in second:
            if s_since > f_till:
                break

            if f_till >= s_since and f_since <= s_till:
                intersects.append((max(f_since, s_since), min(f_till, s_till)))

    return sorted(intersects)


def fit_closest_date(
    now: datetime,
    ranges: TimeRanges,
    tz: Optional[tzinfo] = None,
) -> datetime:
    if ranges is None:
        return now

    previous_tz = now.tzinfo
    result = now if tz is None else now.astimezone(tz)
    current_time = result.time()
    result_time = None
    ranges = sorted(ranges)

    for since, till in ranges:
        if till >= current_time >= since:
            result_time = current_time
            break

        if since > current_time:
            result_time = since
            break

    if result_time is None:
        result += timedelta(days=1)
        result_time = ranges[0][0]

    result = result.replace(
        hour=result_time.hour,
        minute=result_time.minute,
        second=result_time.second,
        microsecond=result_time.microsecond,
    )

    return result.astimezone(previous_tz)


def split_recipients(
    recipients: Iterable[dict],
    now: datetime,
    available: TimeRanges,
    tz_key: str = TZ_KEY,
    immediate_key: str = IMMEDIATE_KEY,
) -> Dict[datetime, List[dict]]:
    to_tz = defaultdict(list)
    result = defaultdict(list)

    for recipient in recipients:
        if recipient.get(immediate_key, False):
            result[now].append(recipient)
            continue

        to_tz[resolve_tz(recipient.get(tz_key, None))].append(recipient)

    tz_to_send_at = {
        tz: fit_closest_date(now, available, tz=tz)
        for tz in to_tz
    }

    for tz, send_at in tz_to_send_at.items():
        result[send_at] += to_tz[tz]

    return result


def split_scheduled_messages(
    channel_availability: List[ChannelAvailability],
    messages: TemplatedMessages,
    now: datetime,
    tz_key: str = TZ_KEY,
    immediate_key: str = IMMEDIATE_KEY,
    events_availability: Dict[str, List[EventAvailability]] = {},
) -> Dict[datetime, TemplatedMessages]:
    to_date = defaultdict(list)
    channel_available = normalize_ranges(
        (x.available_since, x.available_till)
        for x in channel_availability
    )
    event_messages = defaultdict(list)

    for templated_message in messages:
        event_messages[templated_message[0].event].append(templated_message)

    event_available = {
        event: intersect_timeranges(channel_available, normalize_ranges(
            (x.available_since, x.available_till)
            for x in events_availability.get(event, [])
        ))
        for event in event_messages
    }

    for event, messages in event_messages.items():
        available = event_available[event]

        if len(available) == 0:
            to_date[now] += messages
            continue

        for message, template in messages:
            for send_at, recipients in split_recipients(
                message.recipients, now, available, tz_key=tz_key,
                immediate_key=immediate_key,
            ).items():
                to_date[send_at].append((
                    MessageData(
                        channel=message.channel, event=message.event,
                        recipients=recipients, context=message.context,
                    ),
                    template,
                ))

    return to_date



from typing import *
import traceback
import logging
from collections import defaultdict

from django.utils.translation import get_language, override


from wcd_envoyer.events import EventRegistry, registry as events_registry
from wcd_envoyer.channels import (
    FailedMessages,
    BaseMessagingBackend, MessageData, TemplatedMessages, ChannelRegistry,
    registry as channels_registry
)
from wcd_envoyer.const import MessageStatus
from wcd_envoyer.utils import KwargsInjector, sort_and_group
from wcd_envoyer.models import Message, ChannelConfig
from wcd_envoyer.signals import (
    messages_sent, messages_sent_failed, messages_sent_succeeded,
)

from .templates_resolver import TemplatesResolver


__all__ = (
    'TemplatedMessagesGroup', 'MessagesGroup',
    'SucceededMessagesGroup', 'FailedMessagesGroup',
    'Sender',
)

TemplatedMessagesGroup = Tuple[
    BaseMessagingBackend, ChannelConfig, TemplatedMessages
]
MessagesGroup = Tuple[BaseMessagingBackend, ChannelConfig, List[Message]]
SucceededMessagesGroup = Tuple[BaseMessagingBackend, ChannelConfig, List[Message]]
FailedMessagesGroup = Tuple[BaseMessagingBackend, ChannelConfig, FailedMessages]


class Sender(KwargsInjector):
    channels_registry: ChannelRegistry = channels_registry
    events_registry: EventRegistry = events_registry
    templates_resolver: TemplatesResolver = TemplatesResolver()

    def __init__(
        self,
        channels_registry: Optional[ChannelRegistry] = None,
        events_registry: Optional[EventRegistry] = None,
        **kwargs
    ):
        super().__init__(
            channels_registry=channels_registry,
            events_registry=events_registry,
            **kwargs,
        )

    def get_backends(self, only_channels: Optional[Sequence[str]] = None):
        return {
            key: descriptor.backend
            for key, descriptor in self.channels_registry.items()
            if only_channels is None or key in only_channels
        }

    def get_channel_configs(self, channels: Sequence[str]):
        return {
            x.channel: x
            for x in ChannelConfig.objects.filter(channel__in=channels)
        }

    def make_templated_messages(
        self,
        backends: Dict[str, BaseMessagingBackend],
        configs: Dict[str, ChannelConfig],
        messages_data: Sequence[MessageData],
    ) -> List[TemplatedMessagesGroup]:
        templates = self.templates_resolver.get_templates({
            (data.channel, data.event)
            for data in messages_data
            if data.channel in backends
        })

        grouped: Dict[str, TemplatedMessages] = defaultdict(list)

        for data in messages_data:
            template = templates.get((data.channel, data.event))

            if template is not None:
                grouped[data.channel].append((data, template))

        return [
            (backends[channel], configs[channel], data)
            for channel, data in grouped.items()
            if channel in backends and channel in configs
        ]

    def make_backend_messages(
        self,
        backend: BaseMessagingBackend,
        config: ChannelConfig,
        items: TemplatedMessages,
        context: dict = {},
    ) -> List[Message]:
        return backend.make_messages(
            config, items, context=context,
        )

    def make_messages(
        self,
        templated_messages: Iterable[TemplatedMessagesGroup],
        context: dict = {},
    ) -> List[Tuple[BaseMessagingBackend, ChannelConfig, List[Message]]]:
        result = []

        for backend, config, items in templated_messages:
            messages = self.make_backend_messages(
                backend, config, items, context=context,
            )

            if len(messages) > 0:
                result.append((backend, config, messages))

        Message.objects.bulk_create([
            item
            for _, _, items in result
            for item in items
        ])

        return result

    def prepare_messages(
        self,
        messages_data: Sequence[MessageData],
        context: dict = {},
        only_channels: Optional[Sequence[str]] = None,
    ):
        if len(messages_data) == 0:
            return []

        language = context.get('language', None) or get_language()
        context['language'] = language

        with override(language=language):
            backends = self.get_backends(only_channels=only_channels)
            configs = self.get_channel_configs(backends.keys())
            templated_messages = self.make_templated_messages(
                backends, configs, messages_data,
            )
            return self.make_messages(
                templated_messages, context=context,
            )

    def update_messages_status(
        self, status: MessageStatus, messages: Sequence['Message'],
        only_for_status: Optional[MessageStatus] = None, skip_db: bool = False,
    ):
        for message in messages:
            if (
                only_for_status is None
                or
                message.status == only_for_status
            ):
                message.status = status

        if not skip_db:
            Message.objects.bulk_update(messages, fields=['status'])

        return messages

    def send_messages_groups(
        self,
        messages_groups: List[MessagesGroup],
        context: dict = {},
    ) -> Tuple[List[SucceededMessagesGroup], List[FailedMessagesGroup]]:
        # TODO: Too complex and self-repetitive method, should be
        # changed somehow...
        succeeded: List[SucceededMessagesGroup] = []
        failed: List[FailedMessagesGroup] = []

        self.update_messages_status(MessageStatus.PENDING, [
            x
            for _, _, messages in messages_groups
            for x in messages
        ])

        for backend, config, messages in messages_groups:
            try:
                c_succeeded, c_failed = backend.send(
                    config, messages, context=context,
                )
            except Exception as e:
                error = {
                    'exception': e, 'message': str(e),
                    'traceback': traceback.format_exc(),
                }
                c_succeeded = []
                c_failed = [(x, error) for x in messages]

            if len(c_succeeded) > 0:
                succeeded.append((backend, config, c_succeeded))

            if len(c_failed) > 0:
                failed.append((backend, config, c_failed))

        all_succeeded = [y for _, _, x in succeeded for y in x]
        all_failed = [y for _, _, x in failed for y in x]

        self.update_messages_status(
            MessageStatus.SENT,
            # FIXME: WEIRD! But only 1 query instead of 2.
            (
                all_succeeded
                +
                self.update_messages_status(
                    MessageStatus.FAILED, [x for x, _ in all_failed],
                    only_for_status=MessageStatus.PENDING, skip_db=True,
                )
            ),
            only_for_status=MessageStatus.PENDING,
        )

        try:
            if len(succeeded) > 0:
                messages_sent_succeeded.send(
                    self, messages_groups=succeeded, messages=all_succeeded,
                    context=context,
                )

            if len(failed) > 0:
                messages_sent_failed.send(
                    self, messages_groups=failed, messages=all_failed,
                    context=context,
                )

            messages_sent.send(
                self,
                succeeded_messages_groups=succeeded,
                failed_messages_groups=failed,
                messages=all_succeeded + all_failed,
                context=context,
            )
        except Exception as e:
            logging.exception(e)

        return succeeded, failed

    def send_messages(self, messages: Iterable[Message], context: dict = {}):
        backends = self.get_backends()
        configs = self.get_channel_configs(backends.keys())

        return self.send_messages_groups(
            [
                (backends[channel], configs[channel], list(messages))
                for channel, messages in sort_and_group(
                    messages, lambda x: x.channel
                )
                if channel in backends and channel in configs
            ],
            context=context,
        )

    def send(
        self,
        messages_data: Sequence[MessageData],
        context: dict = {},
        only_channels: Optional[Sequence[str]] = None,
    ):
        messages_groups = self.prepare_messages(
            messages_data, context=context, only_channels=only_channels,
        )

        if len(messages_groups) == 0:
            return [], []

        return self.send_messages_groups(messages_groups, context=context)

    def resend(self, messages: Iterable[Message], context: dict = {}):
        return self.send_messages(messages, context=context)



from typing import *
from dataclasses import dataclass, field

from .backend import BaseMessagingBackend
from ..utils import EntryRegistry, cached_import_string


__all__ = 'ChannelInfoType', 'ChannelDescriptor', 'ChannelRegistry',


class ChannelInfoType(TypedDict):
    key: str
    verbose_name: Optional[str]
    backend: Union[str, Type[BaseMessagingBackend]]
    options: Optional[dict]


@dataclass
class ChannelDescriptor:
    key: str
    verbose_name: str
    backend: BaseMessagingBackend

    @classmethod
    def from_info(cls, entry: ChannelInfoType):
        key = entry['key']
        verbose_name = entry.get('verbose_name', None) or key.title()
        backend = entry['backend']
        backend = (
            cached_import_string(backend)
            if isinstance(backend, str) else
            backend
        )

        return cls(
            key=key,
            verbose_name=verbose_name,
            backend=backend(**entry.get('options', {})),
        )


class ChannelRegistry(EntryRegistry):
    def transform_input(self, entry: ChannelInfoType):
        return entry['key'], ChannelDescriptor.from_info(entry)

    def get_choices(self):
        return [(key, x.verbose_name) for key, x in self.items()]



from typing import Any, Dict

from django.template import Context, Template


__all__ = 'django_template_renderer',


def django_template_renderer(template: str, context: Dict[str, Any]) -> str:
    return Template(template).render(Context(context))



from typing import *
from django.utils.translation import pgettext_lazy

from entangled.forms import EntangledModelForm
from entangled import forms as entangled_forms

from wcd_notifications.compat import JSONField

from wcd_envoyer.models import ChannelConfig, Template


__all__ = (
    'BaseConfigForm',
    'BaseTemplateForm',
)

# HACK!
entangled_forms.JSONField = JSONField


class BaseConfigForm(EntangledModelForm):
    class Meta:
        entangled_fields = {'data': []}
        model = ChannelConfig
        untangled_fields = ['channel']


class BaseTemplateForm(EntangledModelForm):
    renderable_fields: List[str] = []

    class Meta:
        entangled_fields = {'data': []}
        model = Template
        untangled_fields = ['channel', 'event', 'is_active']

    def get_renderable_fields(self):
        return self.renderable_fields



from django.dispatch import receiver

from px_domains import Domain
from wcd_envoyer.signals import messages_sent_failed

from .tracker import track_failed_messages_groups


DEFAULT_DOMAIN = Domain('SENDER') | 'ERROR'


@receiver(messages_sent_failed)
def handle_messages_sent_failed(sender, messages_groups, **kwargs):
    track_failed_messages_groups(DEFAULT_DOMAIN, messages_groups)



from django.db import models
from django.utils.translation import pgettext_lazy
from django.conf import settings

from wcd_envoyer.models import Message


__all__ = 'MessageAttachedUser',


class MessageAttachedUser(models.Model):
    class Meta:
        verbose_name = pgettext_lazy('wcd_envoyer', 'Message attached user')
        verbose_name_plural = pgettext_lazy('wcd_envoyer', 'Message attached users')

    message = models.ForeignKey(
        Message, on_delete=models.CASCADE,
        verbose_name=pgettext_lazy('wcd_envoyer', 'Message'),
        related_name='attached_users',
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE,
        verbose_name=pgettext_lazy('wcd_envoyer', 'User'),
        null=False, blank=False, related_name='envoyer_messages',
    )

    def __str__(self) -> str:
        return pgettext_lazy('wcd_envoyer', '#{id}: user #{user_id}').format(
            id=self.message_id, user_id=self.user_id,
        )



from django.contrib import admin

from wcd_envoyer.models import ChannelConfig

from .forms import AddConfigForm
from .utils import FormAdminMixin, JsonDataAdminMixin


__all__ = 'ChannelConfigAdmin',


@admin.register(ChannelConfig)
class ChannelConfigAdmin(JsonDataAdminMixin, FormAdminMixin, admin.ModelAdmin):
    add_form = AddConfigForm
    readonly_fields = 'json_data', 'created_at', 'updated_at'

    def get_backend_form_class(self, request, obj, backend):
        return backend.config_form_class
